import itertools
import logging
from programlib import Program, Language
from typing import Callable, Optional, Iterable, Tuple, List, Generator
import random
import time

from seidr.llm import explore_llm, default_batch_size
from seidr.eval import Evaluation


def standard_ranking(
        candidates: List[Tuple[str, str, list[Callable[[Program], Evaluation]]]],
) -> List[Tuple[str, str, list[Callable[[Program], Evaluation]]]]:
    def avg_score(
            candidate: Tuple[str, str, list[Callable[[Program], Evaluation]]],
    ) -> float:
        prompt, code, evals = candidate
        score = sum(e.score() for e in evals) / len(evals)
        return score

    return sorted(candidates, key=avg_score, reverse=True)


def lexicase_ranking(
        candidates: List[Tuple[str, str, list[Callable[[Program], Evaluation]]]],
) -> Tuple[str, str, list[Callable[[Program], Evaluation]]]:
    pool = [evals for prompt, code, evals in candidates]

    case_count = min(len(evals) for evals in pool)
    cases = list(range(case_count))
    random.shuffle(cases)

    for case_order in itertools.combinations(cases, case_count):
        logging.info(f"Lexicase: test case order {list(reversed(case_order))}")
        # Pseudocode from page 3 of
        # Spector 2012 "Assessment of problem modality by differential performance of lexicase selection in genetic programming: a preliminary report"
        round_winners = range(len(pool))

        # Loop until a single candidate is left
        # Reversing case_order ensures diversity: the first case is always different
        for case in reversed(case_order):
            fitnesses = [pool[idx][case].score() for idx in round_winners]
            logging.info(
                f"Lexicase: "
                f"idx: fitness values on test number {case}"
                f"{[': '.join([str(idx), str(fitness)]) for idx, fitness in zip(round_winners, fitnesses)]}"
            )
            best_fitness = max(fitnesses)

            round_winners = [
                idx
                for idx, fitness in zip(round_winners, fitnesses)
                if fitness == best_fitness
            ]

            logging.info(
                f"Lexicase: "
                f"programs that have max test pass rate of value {best_fitness} on test {case}) {round_winners}"
            )

            if len(round_winners) == 1:
                break

        for idx in round_winners:
            yield candidates[idx]
        for idx in round_winners:
            del candidates[idx]


def beam_search(
        beam: List | Generator,
        update: Callable,
        ranking: Callable = standard_ranking,
        beam_width: int = 100,
) -> Generator:
    """Generic evolutionary algorithm for improving anything"""
    while True:
        parents = []
        for code in beam:
            yield code
            parents.append(code)

        if len(parents) == 0:
            break

        parents = itertools.islice(ranking(parents), beam_width)
        beam = (child for parent in parents for child in update(parent))


def distribute_heat(heat: float, n: int, batch_size: int) -> Tuple[float, float]:
    """Calculate the start temperature `t` and steps `delta_t`
    for increasing temperature of an LLM that will be distributed over the batch."""
    # Setting: we generate `n` outputs in total from an LLM by generating `batch_size` outputs at once,
    # Note that we use `batch_size`=1 for langchain models
    if n == 1:
        t = 0
        delta_t = 0.2
    else:
        # We intentionally avoid temperature=0
        # That would lead to a batch of identical code snippets
        # Update temperature but keep it 1 at max
        batch_count = n // batch_size + 1
        delta_t = heat / batch_count
        t = delta_t

    return t, delta_t


def print_code(code, **vars):
    print(vars)
    print(code)


class SEIDR:
    """Generate, test and debug programs for a set of given problem descriptions and tests"""

    def __init__(
            self,
            task_name: str,
            task_description: str,
            critics: list[Callable[[Program], Evaluation]],
            model_name: str,
            language: str | Language,
            beam_width: int = 10,
            drafts_per_prompt: int = 10,
            explanations_per_program: int = 10,
            repairs_per_explanation: int = 2,
            lexicase_selection: bool = False,
            log_metrics: Callable = print,
            log_attempt: Callable = print_code,
            log_solution: Callable = lambda *args, **kwargs: print('This program is the best!'),
            log_llm_call: Callable = lambda **kwargs: print(kwargs),
            max_programs: Optional[int] = None,
            batch_size: Optional[int] = None,
            ollama_url: Optional[str] = None,
            delay: int = 0  # for Anthropic models
    ) -> None:
        self.task_name = task_name
        self.task_description = task_description
        self.critics = critics
        self.model_name = model_name
        self.language = language
        self.beam_width = beam_width
        self.drafts_per_prompt = drafts_per_prompt
        self.explanations_per_program = explanations_per_program
        self.repairs_per_explanation = repairs_per_explanation
        self.lexicase_selection = lexicase_selection
        self.log_metrics = log_metrics
        self.log_attempt = log_attempt
        self.log_solution = log_solution
        self.log_llm_call = log_llm_call
        self.max_programs = max_programs
        self.ollama_url = ollama_url
        self.delay = delay

        if not batch_size:
            batch_size = default_batch_size(model_name)
        self.batch_size = batch_size

    def draft(self, start_code: str = "") -> Iterable[str]:
        """Create a draft solution with the "generate" prompt template
        that requires first code lines in `start_code`, task description and programming language"""
        batch_size = min(self.batch_size, self.drafts_per_prompt)
        t, delta_t = distribute_heat(1, self.drafts_per_prompt, batch_size)

        return itertools.islice(
            explore_llm(
                t=t,
                delta_t=delta_t,
                mode="generate",
                model_name=self.model_name,
                language=self.language,
                task_name=self.task_name,
                task_description=self.task_description,
                start_code=start_code,
                log_llm_call=self.log_llm_call,
                batch_size=batch_size,
                base_url=self.ollama_url,
            ),
            self.drafts_per_prompt,
        )

    def repair(self, code: str, feedback: str) -> Iterable[str]:
        """Generate `self.explanations_per_program` * `self.repairs_per_explanation` attempts
        to fix the program in `code` so that it passes tests"""
        explain_batch_size = min(self.batch_size, self.explanations_per_program)
        repair_batch_size = min(self.batch_size, self.repairs_per_explanation)

        explain_t, explain_delta_t = distribute_heat(
            1, self.explanations_per_program, self.batch_size
        )
        repair_t, repair_delta_t = distribute_heat(
            1, self.repairs_per_explanation, self.batch_size
        )

        for bug_summary in itertools.islice(
                explore_llm(
                    t=explain_t,
                    delta_t=explain_delta_t,
                    mode="explain_bugs",
                    model_name=self.model_name,
                    language=self.language,
                    task_name=self.task_name,
                    task_description=self.task_description,
                    code=code,
                    issue=feedback,
                    log_llm_call=self.log_llm_call,
                    batch_size=explain_batch_size,
                    base_url=self.ollama_url,
                ),
                self.explanations_per_program,
        ):
            for repair in itertools.islice(
                    explore_llm(
                        t=repair_t,
                        delta_t=repair_delta_t,
                        mode="repair",
                        model_name=self.model_name,
                        language=self.language,
                        task_name=self.task_name,
                        task_description=self.task_description,
                        input=input,
                        code=code,
                        bug_summary=bug_summary,
                        log_llm_call=self.log_llm_call,
                        batch_size=repair_batch_size,
                        base_url=self.ollama_url,
                    ),
                    self.repairs_per_explanation,
            ):
                yield repair

    def develop(self, start_code: str = "") -> str:
        """
        Write a program in a programming language that solves task and passes tests.
        Solve repair-rewrite trade-off with beam search of given beam size

        examples is a sequence of (inputs, outputs) pairs
        where inputs and outputs are sequences of strings (lines of code)
        likewise for tests

        examples are used in the prompt for the language model,
        while tests are used to select the best solution

        Returns a generator of programs where each program passes
        more tests than the previous one. The last program in the generator
        passes all tests.
        """

        def have_kids(
                candidate: Tuple[str, str, list[Callable[[Program], Evaluation]]],
        ) -> Tuple[str, str, list[Callable[[Program], Evaluation]]]:
            prompt, code, evals = candidate
            worst_eval = min(evals, key=lambda e: e.score())
            feedback = worst_eval.pen_report()

            for code in self.repair(code, feedback):
                yield feedback, code, [critic(code) for critic in self.critics]

        drafts = self.draft(start_code)
        drafts = (
            (self.task_description, code, [critic(code) for critic in self.critics])
            for code in drafts
        )

        best_score = float("-inf")
        best_code = None

        ranking = lexicase_ranking if self.lexicase_selection else standard_ranking
        search = beam_search(
            beam=drafts, update=have_kids, ranking=ranking, beam_width=self.beam_width
        )

        for idx, candidate in enumerate(search):
            prompt, code, evals = candidate

            avg_score = sum(e.score() for e in evals) / len(evals)
            test_pass_rate = sum(e.check() for e in evals) / len(evals)

            logging.info(f"Prompt:\n{prompt}\n")
            logging.info(f"The program generated with the prompt above:\n{code}")

            metrics = {"idx": idx, "avg_score": avg_score, "pass_rate": test_pass_rate}

            self.log_metrics(metrics)
            self.log_attempt(
                code, idx=idx, prompt=prompt, test_pass_rate=test_pass_rate
            )

            if avg_score > best_score:
                best_score = avg_score
                best_code = code
                self.log_metrics(
                    {f"best_{metric}": val for metric, val in metrics.items()}
                )
                logging.info(
                    "\n".join(
                        [f"\nbest_{metric}: {val}\n" for metric, val in metrics.items()]
                    )
                )
                self.log_solution(
                    code, idx=idx, prompt=prompt, test_pass_rate=test_pass_rate
                )

                if test_pass_rate == 1:
                    break

            if self.max_programs is not None and (idx == self.max_programs - 1):
                break

            time.sleep(self.delay)

        return best_code



"""
Handy tools for recording our generated programs to a git repository.
May be nice to include this into programlib one day
"""

import logging
import os
import pathlib
import shutil
from pathlib import Path
from uuid import uuid4

import git
import gitdb
from git import GitError
from git import Repo
from tenacity import retry, retry_if_exception_type, stop_after_attempt
from tenacity import wait_random_exponential


@retry(
    retry=retry_if_exception_type(git.exc.GitCommandError)
    | retry_if_exception_type(git.exc.BadName)
    | retry_if_exception_type(gitdb.exc.BadName),
    wait=wait_random_exponential(),
    stop=stop_after_attempt(50),
)
def pullpush(repo):
    repo.remotes.origin.pull()
    repo.remotes.origin.push()


def upload_file(repo: Repo, filename: pathlib.Path | str, message: str = None):
    """Commit and push a file to the repo and log the commit"""
    if not message:
        message = f"added {filename}"

    pullpush(repo)
    repo.index.add(filename)
    repo.index.commit(message)
    pullpush(repo)
    commit_url = repo.remotes.origin.url + "/commit/" + repo.head.commit.hexsha
    commit_url = "https://" + commit_url[commit_url.find("@github.com") + 1 :]
    logging.info(
        f"Tried to push updates to git repo.\n"
        f"See commit at {commit_url}. \n"
        f"Commit message: {message}"
    )


def ensure_repo(remote: str, path: pathlib.Path | str, branch: str = None) -> Repo:
    """Create a remote Git repository at `remote` corresponding to the local `path` and checkout to `branch`"""
    try:
        repo = Repo(path)

        if branch:
            repo.git.checkout(branch)
    except GitError as e:
        logging.info(f"Git error in ensure repo {e}.")

        shutil.rmtree(path, ignore_errors=True)
        repo = Repo.clone_from(remote, path)

        if branch:
            repo.git.fetch("--all")
            branches = [ref.name for ref in repo.references]

            if f"{repo.remote().name}/{branch}" in branches:
                repo.git.checkout(branch)
            else:
                repo.git.checkout("-b" + branch)
                repo.git.add(repo.working_dir)
                repo.git.push("--set-upstream", repo.remote().name, branch)

    return repo


def config_repo(dir: str, branch: str) -> Repo | None:
    """Set up a repository corresponding to the local `dir` with branch name `branch`"""
    try:
        import os

        logging.info("config_repo")
        repo = ensure_repo(os.environ["GIT_REMOTE"], dir, branch=branch)
        repo.config_writer().set_value("user", "name", os.environ["GIT_USER"]).release()
        repo.config_writer().set_value(
            "user", "email", os.environ["GIT_EMAIL"]
        ).release()
        repo.config_writer().set_value("pull", "rebase", False).release()
        return repo
    except (KeyError, GitError) as e:
        logging.info(f"config_repo exception {e}")
        return None


class FileLogger:
    """
    A careful archivist keeping track of your program's versions.
    If git environment variables are set, will synchronize with a remote repository.
    """

    def __init__(self, branch, filename, commit_msg_template="{message}"):
        os.makedirs("solutions", exist_ok=True)
        self.filename = filename.replace("/", "_")
        self.commit_msg_template = commit_msg_template
        cache_dir = branch + "_" + str(uuid4())[:6]
        self.dir = Path("solutions") / cache_dir
        self.repo = config_repo(self.dir, branch=branch)
        os.makedirs(self.dir, exist_ok=True)

    def log(self, content: str, **vars: dict):
        """Log content to file"""
        with open(self.dir / self.filename, "w") as f:
            f.write(content)
        if self.repo:
            commit_msg = self.commit_msg_template.format(**vars)
            upload_file(self.repo, self.filename, commit_msg)

    def __call__(self, *args, **kwargs):
        return self.log(*args, **kwargs)



from string import Template
from typing import List, Tuple

from langchain.prompts import (
    SystemMessagePromptTemplate,
    HumanMessagePromptTemplate,
    ChatPromptTemplate,
)
from programlib import language_

from seidr import get_template


def initial_prompt(task_description: str, examples: Tuple[List[str], List[str]]):
    """Create a docstring for the draft program (to be used in the generate mode)"""
    prompt = task_description
    prompt += "\nFor example,"
    for sample_inputs, sample_outputs in examples:
        prompt += ""
        prompt += "\ninput:\n"
        for sample_input in sample_inputs:
            prompt += sample_input + "\n"
        prompt += "output:"
        for sample_output in sample_outputs:
            prompt += "\n" + sample_output
    return prompt


def start_coding(prompt: str, language: str = "C++") -> str:
    """Fill in the template for the draft program with the first lines"""
    language = language_(language)
    template = language.source.format(name="template")
    template = get_template(template)
    template = Template(template)
    return template.substitute(prompt=prompt)


SYSTEM_PROMPT = """You are an experienced software developer. \
You write concise code in {language}. \
The code must read input from user and return output corresponding to the task description."""

HUMAN_PROMPTS = {
    "generate": "Solve the following code contest problem: {task_name}. "
    "Problem description: {task_description}.\n"
    "{start_code}\n"
    "Only complete the code, do not add triple quotes, do not give explanations.",
    "explain_bugs": "I'm trying to solve the following code contest problem: {task_name}. "
    "Problem description: {task_description}.\n"
    "Currently, the code is \n```\n{code}\n``` \n"
    "The issue is \n"
    "{issue}\n"
    "Describe how I should fix the code in a very concise manner.",
    "repair": "Solve the following code contest problem: {task_name}. "
    "Problem description: {task_description}.\n"
    "Currently, the code is \n```\n{code}\n``` \n"
    "Modify the code as  {bug_summary}.\n"
    "You must only return correct code. Remove any triple quotes, language name or explanations. ",
    "format": "Currently, the code is \n```\n{code}\n``` \n"
    "Keep the current code imports and "
    "add the following header if it is not present: \n{first_lines}\n"
    "Remove triple quotes from the answer. Add input reading from the user."
    "Print output as in the examples, do not print other words.",
}


class ModeNameError(KeyError):
    """Used to catch a wrong human prompt key"""

    pass


def create_chat_prompt_template(
    mode: str = "generate",
) -> ChatPromptTemplate:
    """Returns chat prompt template in the form of two messages, system prompt and human message,
    depending on the `mode`"""
    try:
        system_message_template = SystemMessagePromptTemplate.from_template(
            SYSTEM_PROMPT
        )
        human_message_template = HumanMessagePromptTemplate.from_template(
            HUMAN_PROMPTS[mode]
        )

        chat_prompt_template = ChatPromptTemplate.from_messages(
            [system_message_template, human_message_template]
        )
    except KeyError:
        raise ModeNameError(
            f"Unsupported mode name '{mode}'. Use one of {HUMAN_PROMPTS.keys()}"
        )

    return chat_prompt_template


def ollama_messages(mode: str = "generate", **kwargs) -> List[dict[str, str]]:
    """Returns messages formatted for interaction with ollama"""
    try:
        prompt = [
            {"role": "system", "content": SYSTEM_PROMPT.format(**kwargs)},
            {"role": "user", "content": HUMAN_PROMPTS[mode].format(**kwargs)},
        ]
    except KeyError:
        raise ModeNameError(
            f"Unsupported mode name '{mode}'. Use one of {HUMAN_PROMPTS.keys()}"
        )

    return prompt



from abc import ABC
from programlib import Program

dont_change = "Do not change anything"


class Evaluation(ABC):
    """
    A class for evaluating SEIDR performance by running tests on generated programs (problem solutions)
    Produces a binary pass/fail result, a float score, and a text report
    """

    def __init__(self, SUT, passing_score: float = 1.0):

        """
        SUT: System Under Test
        passing_score: float score required to pass the evaluation
        """

        self.SUT = SUT
        self.passing_score = passing_score

    def check(self) -> bool:
        """Produce a binary pass/fail evaluation"""
        return self.score() >= self.passing_score

    def score(self) -> float:
        """Produce a float score"""
        raise NotImplementedError

    def pen_report(self) -> str:
        """Produce a text report"""
        raise NotImplementedError

    def run_test(self, rerun=True):
        """Test the program on a given I/O pair and update the score, errors or other output.
        Force test by default if `rerun=True`"""
        raise NotImplementedError


class IOMatch(Evaluation):
    def __init__(self, code, language, input, output, task_description=None):
        program = Program(code, language=language)
        super().__init__(program)
        self.input = input
        self.output = output
        self.test_run = None
        self.task_description = task_description

    def run_test(self, rerun=True):
        if rerun or not self.test_run:
            self.test_run = self.SUT.test([[self.input, self.output]])[0]

    def score(self) -> float:
        self.run_test(rerun=False)
        return self.SUT.avg_score

    def pen_report(self) -> str:
        self.run_test(rerun=False)

        if self.check():
            return dont_change
        elif self.test_run.exit_status:
            return "\n".join(self.test_run.output_lines)
        else:
            input = "\n".join(self.test_run.input_lines)
            expected_output = "\n".join(self.test_run.expected_output_lines)
            output = "\n".join(self.test_run.output_lines)
            return f"it must return {expected_output} for input {input}, but it returns {output}. "


class UnitTest(Evaluation):
    def __init__(self, code, language, test):
        program = Program(code + "\n" + test, language=language)
        super().__init__(program)
        self.output = None

    def run_test(self, rerun=True):
        if rerun or not self.output:
            if self.SUT.compile_error:
                self.output = self.SUT.term.emulate(self.SUT.stdout)
            else:
                self.output = self.SUT.run(force=True)

    def score(self) -> float:
        self.run_test(rerun=False)
        return 0.0 if self.SUT.exitstatus else 1.0

    def check(self) -> bool:
        self.run_test(rerun=False)
        return not self.SUT.exitstatus

    def pen_report(self) -> str:
        self.run_test(rerun=False)
        if self.check():
            return dont_change
        else:
            self.output = (
                "\n".join(self.output) if type(self.output) is list else self.output
            )
            return self.output

class Gymnasium(Evaluation):
    def __init__(self, env, code, language, passing_score, error_reward=-1000):
        self.action_mode = type(env.action_space).__name__.lower()
        program = Program(code, language=language)
        super().__init__(program, passing_score)

        self.env = env
        self.tot_reward = 0
        self.tot_txt = ''
        self.done = False
        self.error_reward = error_reward

    def play(self):
        if self.done:
            return

        self.tot_reward = 0
        self.tot_txt = ''
        agent = self.SUT.spawn(action_mode=self.action_mode)

        try:
            observation, info = self.env.reset()
            self.tot_txt += info.get('memos', '')
            terminated = False
            truncated = False
            
            while not (terminated or truncated):
                if 'ascii' in self.env.metadata.get('render.modes', []):
                    ascii_render = self.env.render(mode='ascii')
                    self.tot_txt += ascii_render

                action, _ = agent.predict(observation, deterministic=True)

                observation, reward, terminated, truncated, info = self.env.step(action)
                self.tot_reward += reward
                self.tot_txt += info.get('memos', '')
        except RuntimeError as e:
            self.tot_reward = self.error_reward
            self.tot_txt += f'FATAL {e}'
        finally:
            agent.close()

        self.done = True

    def score(self):
        self.play()
        return self.tot_reward

    def pen_report(self):
        self.play()
        self.tot_txt += f'\nFinal reward: {self.tot_reward}'
        return self.tot_txt


import logging
import os

from langchain_community.chat_models import ChatOpenAI, ChatOllama
from langchain_anthropic import ChatAnthropic
from typing import Optional, Callable, Iterable
from black import format_str, FileMode
from langchain_core.runnables import RunnableSequence
from pytest_codeblocks import extract_from_buffer
from io import StringIO

from programlib import Language, language_
from seidr.prompt import create_chat_prompt_template

token_error_message = (
    "tokens for the input and instruction but the maximum allowed is 3000. "
    "Please reduce the input or instruction length."
)


def extract_codes(message_content: str, language: Language | str) -> str:
    """Extract code out of a message and (if Python) format it with black"""

    try:
        code_blocks = list(extract_from_buffer(StringIO(message_content)))
        code_blocks = [code for code in code_blocks if bool(code)]
    except RuntimeError as e:
        code_blocks = []

    if not code_blocks:
        yield message_content
    else:
        for code_block in code_blocks:
            if language_(language).name == "Python":
                yield run_black(code_block.code).strip()
            else:
                yield code_block.code.strip()


def run_black(code: str) -> str:
    """Format (Python) code with Black"""
    try:
        return format_str(code, mode=FileMode())
    except Exception as e:
        logging.info(e)
        return code

def which_api(model_name):
    model_name = model_name.lower()
    if "gpt" in model_name or "deepseek" in model_name:
        return ChatOpenAI
    elif "claude" in model_name:
        return ChatAnthropic
    else:
        return ChatOllama
    
def default_batch_size(model_name: str):
    """Limit batch size to 1 for Ollama, because it does not support batching"""
    if which_api(model_name) == ChatOllama:
        return 1
    else:
        return 10


def create_chain(
    temperature: float = 0.0,
    mode: str = "generate",
    model_name: str = "codellama:7b-instruct",
    base_url: Optional[str] = None,
) -> RunnableSequence:
    """Set up a LangChain pipeline"""
    chat_prompt_template = create_chat_prompt_template(mode)
    api = which_api(model_name)

    if api == ChatOpenAI:
        chat_model = ChatOpenAI(
            model=model_name,
            temperature=temperature,
            openai_api_base=os.getenv("OPENAI_API_BASE"),
            openai_api_key=os.getenv("OPENAI_API_KEY"),
            openai_organization=os.getenv("OPENAI_ORG"),
        )
    elif api == ChatAnthropic:
        chat_model = ChatAnthropic(
            model_name=model_name,
            temperature=temperature,
            anthropic_api_key=os.getenv('ANTHROPIC_API_KEY')
        )
    elif api == ChatOllama:
        chat_model = ChatOllama(
            base_url=base_url, model=model_name, temperature=temperature
        )

    return chat_prompt_template | chat_model


def query_llm(
    language: Language | str,
    base_url: str,
    temperature: float = 0.0,
    mode: str = "generate",
    model_name: str = "codellama:7b-instruct",
    n: int = 1,
    **kwargs,
) -> list[str]:
    """Generate `n` outputs with an LLM"""
    logging.info(
        f"Query LLM ({model_name}) in mode {mode} with temperature {temperature}\n"
    )

    kwargs["language"] = str(language)

    # if "gpt" in model_name.lower():
    chain = create_chain(
        temperature=temperature, mode=mode, model_name=model_name, base_url=base_url
    )
    result = chain.batch([kwargs for _ in range(n)])
    result = [r.content for r in result]

    if mode == "repair":
        logging.info(
            f"Generating repair candidates for bug summary: \n{kwargs['bug_summary']}\n"
        )
    elif mode == "explain_bugs":
        logging.info(f"Generating explanations for code: \n{kwargs['code']}\n")

    if mode != "explain_bugs":
        result = [
            c
            for r in result
            for c in extract_codes(message_content=r, language=language)
        ]
        result_logging = "\n\n".join(result)
        logging.info(f"LLM output after code extraction: \n{result_logging}\n")
    else:
        result_logging = "\n\n".join(result)
        logging.info(f"LLM output: \n{result_logging}\n")

    return result


def explore_llm(
    language: Language | str,
    base_url: str,
    log_llm_call: Callable = lambda **kwargs: None,
    mode: str = "generate",
    model_name: str = "codellama:7b-instruct",
    t: float = 0.0,
    delta_t: float = 0.2,
    batch_size: int = 1,
    **kwargs,
) -> Iterable[str]:
    """Generate LLM outputs and increase temperature for every new batch"""
    while t <= 1:
        log_llm_call(**locals())
        yield from query_llm(
            language=language,
            temperature=t,
            mode=mode,
            model_name=model_name,
            n=batch_size,
            base_url=base_url,
            **kwargs,
        )

        t += delta_t


if __name__ == "__main__":
    import itertools

    logging.basicConfig(level=logging.INFO)

    fizz_buzz = """
    Write a Python program that iterates integer numbers
    from 1 to 50. For multiples of three print "Fizz"
    instead of the number and for the multiples of five
    print "Buzz". For numbers which are multiples of both
    three and five print "FizzBuzz".
    """

    for model_name in ["codellama:34b-instruct", "gpt-3.5-turbo"]:
        for code in itertools.islice(
            explore_llm(
                language="Python",
                task_name="fizz-buzz",
                task_description=fizz_buzz,
                start_code="",
                model_name=model_name,
            ),
            10,
        ):
            print(code)
